"""Contains the TimeseriesExtractor class for extracting timeseries"""

import json, os, re
from functools import lru_cache
from typing import Any, Callable, Literal, Optional, Union
from typing_extensions import Self

import matplotlib.pyplot as plt, numpy as np
from joblib import Parallel, delayed
from numpy.typing import NDArray
from pandas import DataFrame
from tqdm.auto import tqdm

import neurocaps._utils.io as io_utils
from ..exceptions import BIDSQueryError
from ..typing import ParcelConfig, ParcelApproach, SubjectTimeseries
from .._utils import (
    _TimeseriesExtractorGetter,
    _PlotDefaults,
    _PlotFuncs,
    _check_kwargs,
    _check_confound_names,
    _check_parcel_approach,
    _extract_timeseries,
    _logger,
)

LG = _logger(__name__)


class TimeseriesExtractor(_TimeseriesExtractorGetter):
    """
    Timeseries Extractor Class.

    Performs timeseries denoising, extraction, serialization (pickling), and visualization.

    Parameters
    ----------
    space: :obj:`str`, default="MNI152NLin2009cAsym"
        The standard template space that the preprocessed bold data is registered to.

    parcel_approach: :obj:`ParcelConfig`, :obj:`ParcelApproach`, :obj:`str`, or :obj:`None`, default=None
        Specifies the parcellation approach to use. Options are "Schaefer", "AAL", or "Custom". Can
        be initialized with parameters, as a nested dictionary, or loaded from a pickle file. For
        detailed documentation on the expected structure, see the type definitions for
        ``ParcelConfig`` and ``ParcelApproach`` in the "See Also" section. Defaults to
        ``{"Schaefer": {"n_rois": 400, "yeo_networks": 7, "resolution_mm": 1}}`` if None.

        .. versionchanged:: 0.28.0 Default changed to None; however, the same default dictionary\
        configuration remains the same and is backwards compatible.

    standardize: :obj:`bool` or or :obj:`None`, default=True
        Standardizes the timeseries (zero mean and unit variance using sample standard deviation).
        Always the final step in the pipeline.

        .. versionchanged:: 0.25.0 No longer passed to Nilearn's ``NiftiLabelsMasker`` and only\
        performs standardization using sample standard deviation. Default behavior of standardizing\
        using sample standard deviation is the same; however, when not None or False, standardizing\
        is always done at the end of the pipeline to prevent any external standardization from\
        needing to be done when censoring or extracting condition.

        .. note:: Standard deviations below ``np.finfo(std.dtype).eps`` are replaced with 1 for\
        numerical stability.

    detrend: :obj:`bool`, default=False
        Detrends the timeseries.

        .. versionchanged:: 0.26.0 Default changed from True to False due to the redundancy of\
        detrending when discrete cosine-basis regressors are used, which is included in the "basic"\
        option for ``confound_names``.

    low_pass: :obj:`float`, :obj:`int`, or :obj:`None`, default=None
        Filters out signals above the specified cutoff frequency.

    high_pass: :obj:`float`, :obj:`int`, or :obj:`None``, default=None
        Filters out signals below the specified cutoff frequency.

    fwhm: :obj:`float`, :obj:`int`, or :obj:`None`, default=None
        Applies spatial smoothing to data (in millimeters).

    use_confounds: :obj:`bool`, default=True
        If True, performs nuisance regression during timeseries extraction using the default or
        user-specified confounds in ``confound_names``.

        .. important:: Requires the confound tsv files to be in same directory as preprocessed BOLD
        images.

    confound_names: {"basic"}, :obj:`list[str]`, or :obj:`None`, default="basic"
        Names of confounds extracted from the confound tsv files if ``use_confounds=True``.

        If "basic", the following confounds are used by default:

        - All discrete cosine-basis regressors.
        - Six head-motion parameters and their first-order derivatives.
        - First six principal aCompCor components.

        .. important::
            - Confound names follow fMRIPrep's naming scheme (versions >= 1.2.0).
            - Wildcards are supported: e.g., "cosine*" matches all confounds starting with "cosine".

    fd_threshold: :obj:`float`, :obj:`dict[str, float | int]`, or :obj:`None`, default=None
        Threshold for volume censoring based on framewise displacement (FD). Computed only after
        dummy volumes are removed.

        - *If float*, removes volumes where FD > threshold.
        - *If dict*, the following subkeys are available **(all non-required subkeys are None by default)**:

            - "threshold": A float (required). Removes volumes where FD > threshold.
            - "outlier_percentage": A float in interval [0,1]. Removes entire runs where proportion\
              of censored volumes exceeds this threshold. Proportion calculated after dummy scan removal.

            .. note::
                - A warning is issued when a run is flagged.
                - If ``condition`` specified for task-based data in ``self.get_bold()``, only\
                  considers volumes associated with the condition.

            - "n_before": An integer. Indicates the number of volumes to remove before each flagged volume.
            - "n_after": An integer. Indicates the number of volumes to remove after each flagged volume.
            - "use_sample_mask": A boolean. Controls when censoring is applied in the processing pipeline.

            .. important::
                - When True:

                    - Passes a sample mask of censored volumes to Nilearn's ``NiftiLabelsMasker``.
                    - Sets ``clean__extrapolate=False`` to prevent interpolation of end volumes.
                    - Censoring is applied before nuisance regression.
                    - If ``condition`` is specified for task-based data in ``self.get_bold()``, the\
                      timeseries is temporarily padded to extract the correct frames.

                - When False or None:

                    - Full timeseries data is used during nuisance regression.
                    - Censoring is applied after nuisance regression.

            - "interpolate": A boolean. If True, uses scipy's ``CubicSpline`` function` to perform\
              cubic spline interpolation only on censored frames. **Only performs interpolation if True**.

            .. note:: Interpolation is only performed on frames that are bounded by non-censored\
            frames on both ends. For example, given a ``censor_mask=[0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0]``\
            where "0" indicates censored high motion volumes and "1" indicates non-censored, low\
            motion volumes, only the volumes at index 3, 5, 6, 7, and 9 would be interpolated.

        .. important::
            - A column named "framewise_displacement" must be available in the confounds file.
            - ``use_confounds`` must be set to True.
            - Do not specify "framewise_displacement" in ``confound_names``.
            - See Nilearn's documentation for details on censored volume handling when\
            "use_sample_mask" is True:

                - `Signal Clean <https://nilearn.github.io/stable/modules/generated/nilearn.signal.clean.html>`_
                - `NiftiLabelsMasker \<https://nilearn.github.io/stable/modules/generated/nilearn.maskers.NiftiLabelsMasker.html>`_

            - If "interpolate" is True, then interpolation is only applied after the nuisance\
              regression and parcellation steps have been completed.
            - See Scipy's `CubicSpline <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.CubicSpline.html>`_\
              documentation.

    n_acompcor_separate: :obj:`int` or :obj:`None`, default=None
        Number of aCompCor components to extract separately from the white-matter (WM) and CSF
        masks. Uses first "n" components from each mask separately. For instance, if
        ``n_acompcor_separate=5``, then the the first 5 WM components and the first 5 CSF components
        (totaling 10 components) are regressed out.

        .. important::
            - ``use_confounds`` must be set to True.
            - If specified, this parameter overrides any aCompCor components listed in\
              ``confound_names``.

    dummy_scans: :obj:`int`, :obj:`dict[str, bool | int]`, "auto", or :obj:`None`, default=None
        Number of initial volumes to remove before timeseries extraction.

        - *If int*, removes first "n" volumes.
        - *If auto*, removes first "n" volumes based on "non_steady_state_outlier_XX" columns.
        - *If dict*, the following keys are supported **(all non-required subkeys are None by default)**:

            - "auto": A boolean (required). Removes first "n" volumes based on\
              "non_steady_state_outlier_XX" columns.
            - "min": An integer. Minimum volumes to remove when auto is set to True. If "auto"\
              finds 2 outliers but ``{"min": 3}``, removes 3 volumes.
            - "max": An integer. Maximum volumes to remove when auto is set to True. If "auto"\
              finds 6 outliers but ``{"max": 5}``, removes 5 volumes.

        .. important::
            - "auto" and dictionary option requires ``use_confounds`` to be True and\
              "non_steady_state_outlier_XX" to be present in the confounds tsv file.
            - "min" and "max" keys only apply when "auto" is True.

    dtype: :obj:`str` or "auto", default=None
        The NumPy dtype to convert NIfTI images to.


    Properties
    ----------
    space: :obj:`str`
        The standard template space that the preprocessed BOLD data is registered to. This property
        is also settable.

    parcel_approach: :obj:`ParcelApproach`
        Parcellation information with "maps" (path to parcellation file), "nodes" (labels), and
        "regions" (anatomical regions or networks). Returns a deep copy.

    signal_clean_info: :obj:`dict[str, bool | int | float | str]` or :obj:`None`
        Dictionary containing signal cleaning parameters. Returns a deep copy.

    task_info: :obj:`dict[str, str | int]` or :obj:`None`
        Dictionary containing all task-related information such. Defined after running
        ``self.get_bold()``.

    subject_ids: :obj:`list[str]` or :obj:`None`
        A list containing all subject IDs that have retrieved from PyBIDS and subjected to
        timeseries extraction. Defined after running ``self.get_bold()``.

    n_cores: :obj:`int` or :obj:`None`
        Number of cores used for multiprocessing with Joblib. Defined after running
        ``self.get_bold()``.

    subject_timeseries: :obj:`SubjectTimeseries` or :obj:`None`
        A dictionary mapping subject IDs to their run IDs and their associated timeseries
        (TRs x ROIs) as a NumPy array. Can be deleted using ``del self.subject_timeseries``.
        Defined after running ``self.get_bold()``. This property is also settable (accepts a
        dictionary or pickle file).

    qc: :obj:`dict` or :obj:`None`
        A dictionary reporting quality control, which maps subject IDs to their run IDs and
        information related to framewise displacement and dummy scans.

        ::

            {"subjectID": {"run-ID": {"mean_fd": float, "std_fd": float, ...}}}

    See Also
    --------
    :class:`neurocaps.typing.ParcelConfig`
        Type definition representing the configuration options and structure for the Schaefer and
        AAL parcellations.
    :class:`neurocaps.typing.ParcelApproach`
        Type definition representing the structure of the Schaefer, AAL, and Custom parcellation
        approaches
    :data:`neurocaps.typing.SubjectTimeseries`
        Type definition representing the structure of the subject timeseries.
    Note
    ----
    **Passed Parameters**: ``detrend``, ``low_pass``, ``high_pass``, ``fwhm``, and nuisance
    regressors (``confound_names``) uses ``nilearn.maskers.NiftiLabelsMasker``. The ``dtype``
    parameter is used by ``nilearn.image.load_img``.

    **Custom Parcellations**: Refer to the `NeuroCAPs' Parcellation Documentation
    <https://neurocaps.readthedocs.io/en/stable/parcellations.html>`_ for detailed explanations and
    example structures for Custom parcellations.
    """

    def __init__(
        self,
        space: str = "MNI152NLin2009cAsym",
        parcel_approach: Union[ParcelConfig, ParcelApproach, str, None] = None,
        standardize: bool = True,
        detrend: bool = False,
        low_pass: Optional[Union[float, int]] = None,
        high_pass: Optional[Union[float, int]] = None,
        fwhm: Optional[Union[float, int]] = None,
        use_confounds: bool = True,
        confound_names: Optional[Union[list[str], Literal["basic"]]] = "basic",
        fd_threshold: Optional[Union[float, dict[str, Union[bool, float, int]]]] = None,
        n_acompcor_separate: Optional[int] = None,
        dummy_scans: Optional[Union[int, dict[str, Union[bool, int]], Literal["auto"]]] = None,
        dtype: Optional[Union[str, Literal["auto"]]] = None,
    ) -> None:

        self._space = space
        # Check parcel_approach
        self._parcel_approach = _check_parcel_approach(parcel_approach=parcel_approach)

        if use_confounds:
            if confound_names:
                # Replace confounds if not None
                confound_names = _check_confound_names(
                    high_pass, confound_names, n_acompcor_separate
                )

            if fd_threshold:
                self._validate_init_params("fd_threshold", fd_threshold)

            if dummy_scans:
                self._validate_init_params("dummy_scans", dummy_scans)
        else:
            if confound_names:
                LG.warning(
                    "`confound_names` is specified but `use_confounds` is not True so nuisance "
                    "regression will not be done."
                )
                confound_names = None

            if fd_threshold:
                raise ValueError(
                    "`fd_threshold` specified but `use_confounds` is not True, so removal of "
                    "volumes after nuisance regression cannot be done since confounds tsv file "
                    "generated by fMRIPrep is needed."
                )

            if dummy_scans == "auto" or (isinstance(dummy_scans, dict) and dummy_scans.get("auto")):
                raise ValueError(
                    "'auto' specified in `dummy_scans` but `use_confounds` is not True, so "
                    "automated dummy scans detection cannot be done since confounds tsv file "
                    "generated by fMRIPrep is needed."
                )

            if n_acompcor_separate:
                raise ValueError(
                    "`n_acompcor_separate` specified and `use_confounds` is not True, so separate "
                    "WM and CSF components cannot be regressed out since confounds tsv file "
                    "generated by fMRIPrep is needed."
                )

        self._signal_clean_info = {
            "masker_init": {
                "detrend": detrend,
                "low_pass": low_pass,
                "high_pass": high_pass,
                "smoothing_fwhm": fwhm,
            },
            "standardize": standardize,
            "use_confounds": use_confounds,
            "confound_names": confound_names,
            "n_acompcor_separate": n_acompcor_separate,
            "dummy_scans": dummy_scans.copy() if isinstance(dummy_scans, dict) else dummy_scans,
            "fd_threshold": fd_threshold.copy() if isinstance(fd_threshold, dict) else fd_threshold,
            "dtype": dtype,
        }

    @staticmethod
    def _validate_init_params(param: str, struct: Any) -> None:
        """Validates structure for ``dummy_scans`` and ``fd_threshold``."""
        mandatory_keys = {
            "dummy_scans": {"auto": (bool,)},
            "fd_threshold": {"threshold": (float, int)},
        }

        optional_keys = {
            "dummy_scans": {"min": (int,), "max": (int,)},
            "fd_threshold": {
                "n_before": (int,),
                "n_after": (int,),
                "outlier_percentage": (float,),
                "use_sample_mask": (bool,),
                "interpolate": (bool,),
            },
        }

        type2str = lambda t: t.__name__
        str2text = lambda valid_types: ", ".join([type2str(t) for t in valid_types])

        valid_types = (dict, int, str) if param == "dummy_scans" else (dict, float, int)
        if not isinstance(struct, valid_types):
            raise TypeError(
                f"`{param}` must be one of the following types when not None: "
                f"{str2text(valid_types)}."
            )

        if param == "dummy_scans" and (isinstance(struct, str) and not struct == "auto"):
            raise ValueError(f"'auto' is the only valid string for `dummy_scans`.")

        if isinstance(struct, dict):
            # Check mandatory keys
            key = list(mandatory_keys[param].keys())[0]
            if key not in struct:
                raise KeyError(f"'{key}' is a mandatory key when `{param}` is a dictionary.")
            if not isinstance(struct[key], mandatory_keys[param][key]):
                raise TypeError(
                    f"'{key}' must be of one of the following types: "
                    f"{str2text(mandatory_keys[param][key])}."
                )

            # Check optional keys
            for key in optional_keys[param]:
                if key in struct:
                    if struct[key] is not None and not isinstance(
                        struct[key], optional_keys[param][key]
                    ):
                        raise TypeError(
                            f"'{key}' must be either None or of type "
                            f"{str2text(optional_keys[param][key])}."
                        )
                    # Additional check for "outlier_percentage"
                    if key == "outlier_percentage" and not 0 < struct["outlier_percentage"] < 1:
                        raise ValueError(
                            "'outlier_percentage' must be either None or a float between 0 and 1."
                        )

            # Check invalid keys
            set_diff = set(struct) - set(optional_keys[param]) - set(mandatory_keys[param])
            if set_diff:
                formatted_string = ", ".join(["'{a}'".format(a=x) for x in set_diff])
                LG.warning(
                    f"The following invalid keys have been found in `{param}` and will be ignored: "
                    f"{formatted_string}."
                )

    def get_bold(
        self,
        bids_dir: str,
        task: str,
        session: Optional[Union[int, str]] = None,
        runs: Optional[Union[int, str, list[int], list[str]]] = None,
        condition: Optional[str] = None,
        condition_tr_shift: int = 0,
        tr: Optional[Union[int, float]] = None,
        slice_time_ref: Union[int, float] = 0.0,
        run_subjects: Optional[Union[str, list[str]]] = None,
        exclude_subjects: Optional[Union[str, list[str]]] = None,
        exclude_niftis: Optional[Union[str, list[str]]] = None,
        pipeline_name: Optional[str] = None,
        n_cores: Optional[int] = None,
        parallel_log_config: Optional[dict[str, Union[Callable, int]]] = None,
        verbose: bool = True,
        flush: bool = False,
        progress_bar: bool = False,
    ) -> Self:
        """
        Retrieve Preprocessed BOLD Data from BIDS Datasets.

        Extracts the timeseries data from preprocessed BOLD images located in the derivatives folder
        of a BIDS-compliant dataset. The timeseries data of all subjects are appended to a single
        dictionary ``self.subject_timeseries``.

        .. important::
            - For proper querying, a "dataset_description.json" file must be located in the root of
              the BIDs directory and the pipeline directory (located in the derivatives folder).
            - Refer to `NeuroCAPs' BIDS Structure and Entities documentation\
              <https://neurocaps.readthedocs.io/en/stable/bids.html>`_ for additional information
              about the expected directory structure and file naming scheme (entities) needed for
              querying.
            - This pipeline is most optimized for BOLD data preprocessed by fMRIPrep.

        Parameters
        ----------
        bids_dir: :obj:`str`
            Path to a BIDS compliant directory. A "dataset_description.json" file must be located
            in this directory or an error will be raised.

        task: :obj:`str`
            Name of task to extract timeseries data from (i.e "rest", "n-back", etc).

        session: :obj:`int`, :obj:`str`, or :obj:`None`, default=None
            The session ID to extract timeseries data from. Only a single session can be extracted
            at a time. The value can be an integer (e.g. ``session=2``) or a string (e.g.
            ``session="001"``).

        runs: :obj:`int`, :obj:`str`, :obj:`list[int]`, :obj:`list[str]`, or :obj:`None`, default=None
            List of run numbers to extract timeseries data from (e.g. ``runs=["000", "001"]``).
            Extracts all runs if unspecified.

        condition: :obj:`str` or :obj:`None`, default=None
            Isolates the timeseries data corresponding to a specific condition (listed in the
            "trial_type" column of the "events.tsv" file) after the timeseries has been extracted
            and subjected to nuisance regression. Only a single condition can be extracted at a time.

        condition_tr_shift: :obj:`int`, default=0
            Number of TR units to units to offset both the start and end scan indices of a condition
            to account for a fixed hemodynamic delay. This parameter only applies when a ``condition``
            is specified. For more details about how this offset affects the calculation of task
            conditions, see the "Extraction of Task Conditions" section below.

        tr: :obj:`int`, :obj:`float` or :obj:`None`, default=None
            Repetition time (TR), in seconds, for the specified task. If not provided, the TR will
            be automatically extracted from the first BOLD metadata file found for the task,
            searching first in the pipeline directory, then in the ``bids_dir`` if not found.

        slice_time_ref: :obj:`int` or :obj:`float`, default=0.0
            The reference slice expressed as a fraction of the ``tr`` that is subtracted from the
            condition onset times to adjust for slice time correction when ``condition`` is not None.
            Values can range from 0 to 1. For more details, see the "Extraction of Task Conditions"
            section below.

        run_subjects: :obj:`str`, :obj:`list[str]` or :obj:`None`, default=None
            A string (if single subject) or list of subject IDs to process (e.g.
            ``run_subjects=["01", "02"]``). Processes all subjects if None.

        exclude_subjects: :obj:`str`, :obj:`list[str]` or :obj:`None`, default=None
            A string (if single subject) or list of subject IDs to exclude (e.g.
            ``exclude_subjects=["01", "02"]``).

        exclude_niftis: :obj:`str`, :obj:`list[str]` or :obj:`None`, default=None
            A string (if single file) or List of the specific preprocessed NIfTI files to exclude,
            preventing their timeseries data from being extracted. Used if there are specific runs
            across different participants that need to be excluded.

        pipeline_name: :obj:`str` or :obj:`None`, default=None
            The name of the pipeline folder in the derivatives folder containing the preprocessed
            data. Used if multiple pipeline folders exist in the derivatives folder or the pipeline
            folder is nested (e.g. "fmriprep/fmriprep-20.0.0").

        n_cores: :obj:`int` or :obj:`None`, default=None
            The number of cores to use for multiprocessing with Joblib. The "loky" backend is used.

        parallel_log_config: :obj:`dict[str, multiprocessing.Manager.Queue | int]`
            Passes a user-defined managed queue and logging level to the internal timeseries
            extraction function when parallel processing (``n_cores``) is used. Available dictionary
            keys are:

            - "queue": The instance of ``multiprocessing.Manager.Queue`` to pass to ``QueueHandler``.
              If not specified, all logs will output to ``sys.stdout``.
            - "level": The logging level (e.g. ``logging.INFO``, ``logging.WARNING``). If not
              specified, the default level is ``logging.INFO``.

            Refer to the `NeuroCAPs' Logging documentation\
            <https://neurocaps.readthedocs.io/en/stable/logging.html>`_ for a detailed example
            setting up this parameter.

        verbose: :obj:`bool`, default=True
            If True, logs detailed subject-specific information including: subjects skipped due to
            missing required files, current subject being processed for timeseries extraction,
            confounds identified for nuisance regression in addition to requested confounds that are
            missing for a subject, and additional warnings encountered during the timeseries
            extraction process.

        flush: :obj:`bool`, default=False
            If True, flushes the logged subject-specific information produced during the timeseries
            extraction process.

        progress_bar: :obj:`bool`, default=False
            If True, displays a progress bar.

        Returns
        -------
        self

        Raises
        ------
        BIDSQueryError
            Occurs when subject IDs were not found during querying.

        See Also
        --------
        :data:`neurocaps.typing.SubjectTimeseries`
            Type definition representing the structure of the subject timeseries.

        Important
        ---------
        **Subject Timeseries Dictionary**: This function stores the extracted timeseries of all
        subjects in the ``subject_timeseries`` property and can be deleted using
        ``del self.subject_timeseries`` (Note that ``self.timeseries_to_pickle()`` and
        ``self.visualize_bold()`` need this property in order to be used). The structure is a
        dictionary mapping subject IDs to their run IDs and their associated timeseries (TRs x ROIs)
        as NumPy array. Refer to documentation for ``SubjectTimeseries`` in the "See Also" section
        for an example structure.

        **Data/Property Persistence**: Each time this function is called, it's associated properties
        such as ``self.subject_timeseries``, ``self.task_info``, ``self.qc``, etc, are automatically
        initialized/overwritten to create a clean state for the subsequent analysis. To save, the
        subject timeseries dictionary, ``self.timeseries_to_pickle()`` can be used. Additionally, to
        save the quality control dictionary, ``self.report_qc()`` can be used.

        **NifTI Files Without "run-" Entity**: By default, "run-0" will be used as a placeholder,
        if run IDs are not specified in the NifTI file.

        **Parcellation & Nuisance Regression**: For timeseries extraction, nuisance regression, and
        spatial dimensionality reduction using a parcellation, Nilearn's ``NiftiLabelsMasker``
        function is used. If requested, dummy scans are removed from the NIfTI images and confound
        dataset prior to timeseries extraction. For volumes exceeding a specified framewise
        displacement (FD) threshold, if the "use_sample_mask" key in the ``fd_threshold``
        dictionary is set to True, then a boolean sample mask is generated (where False indicates
        the high motion volumes) and passed to the ``sample_mask`` parameter in Nilearn's
        ``NiftiLabelsMasker``. If, "use_sample_mask" key is False or not specified in the
        ``fd_threshold`` dictionary, then censoring is done after nuisance regression, which is the
        default behavior.

        **Extraction of Task Conditions**: The formula used for computing the scan indices
        corresponding to the corresponding to a specific condition:

        ::

            adjusted_onset = condition_df.loc[i, "onset"] - slice_time_ref * tr
            onset_scan = math.floor(adjusted_onset / tr)
            onset_scan += condition_tr_shift
            end_scan = math.ceil((adjusted_onset + condition_df.loc[i, "duration"]) / tr)
            end_scan += condition_tr_shift

            onset_scan = max([0, onset_scan])
            end_scan = max([0, end_scan])
            scans.extend(range(onset_scan, end_scan))
            scans = sorted(list(set(scans)))

        .. versionchanged:: 0.28.4 Max check done for ``onset_scan`` and ``end_scan`` instead of\
        ``adjusted_onset``.

        When partial scans are computed, ``math.floor`` is used to round down for the beginning scan
        index and ``math.ceil`` is used to round up for the ending scan index. Negative scan indices
        are set to 0 to avoid unintentional negative indexing. For simplicity, note that when
        ``slice_time_ref`` and ``condition_tr_shift`` are 0, the formula simplifies to:

        ::

            start_scan = math.floor(onset / tr)
            end_scan = math.ceil((onset + duration) / tr)
            scans.extend(range(onset_scan, end_scan))
            scans = sorted(list(set(scans)))

        Filtering a specific condition from the timeseries is done after nuisance regression. The
        indices corresponding to the condition are used to extract the TRs (the timepoints that
        fall within the the event window(s) adjusted by the slice timing reference
        (``slice_time_ref``) and a fixed hemodynamic delay (``condition_tr_shift``) if specified)
        from the timeseries.

        If the "use_sample_mask" key in the ``fd_threshold`` dictionary is set to True, the
        truncated 2D timeseries is temporarily padded to ensure the correct rows corresponding to
        the condition are obtained.

        If the "interpolate" key in the ``fd_threshold`` dictionary is set to True, interpolation is
        performed using the full timeseries data (excluding dummy volumes) to replace only the
        censored (high-motion) volumes. Then, the indices corresponding to the condition are
        extracted from the timeseries, excluding any frames that do not have non-censored data at
        both edges.
        """
        if runs and not isinstance(runs, list):
            runs = [runs]

        # Update attributes
        self._task_info = {
            "task": task,
            "session": session,
            "runs": runs,
            "condition": condition,
            "condition_tr_shift": condition_tr_shift,
            "tr": tr,
            "slice_time_ref": slice_time_ref,
        }

        # Quick check condition_tr_shift and slice_time_ref to not cause issues later in the pipeline
        self._validate_get_bold_params()

        # Initialize new attributes
        self._subject_ids = []
        self._subject_timeseries = {}
        self._qc = {}
        self._subject_info = {}
        self._n_cores = n_cores

        layout = self._call_layout(bids_dir, pipeline_name)
        subj_ids = sorted(
            layout.get(
                return_type="id", target="subject", task=task, space=self._space, suffix="bold"
            )
        )

        if not subj_ids:
            msg = (
                "No subject IDs found - potential reasons:\n"
                "1. Incorrect template space (default: 'MNI152NLin2009cAsym'). "
                "Fix: Set correct template space using "
                "`self.space = 'TEMPLATE_SPACE'` (e.g. 'MNI152NLin6Asym')\n"
                "2. File names do not contain specific entities required for "
                "querying such as 'sub-', 'space-', 'task-', or 'desc-' "
                "(e.g 'sub-01_ses-1_task-rest_space-MNI152NLin2009cAsym_desc-preproc-bold.nii.gz')\n"
                "3. Incorrect task name specified in `task` parameter.\n"
                "4. The cache may need to be cleared using "
                "``TimeseriesExtractor._call_layout.cache_clear()`` if the directory has been "
                "changed (e.g. new files added, file names changed, etc) during the current Python "
                "session."
            )
            raise BIDSQueryError(msg)

        if exclude_subjects:
            exclude_subjects = (
                exclude_subjects if isinstance(exclude_subjects, list) else [exclude_subjects]
            )
            exclude_subjects = [
                subj_id.removeprefix("sub-") for subj_id in map(str, exclude_subjects)
            ]
            subj_ids = sorted([subj_id for subj_id in subj_ids if subj_id not in exclude_subjects])

        if run_subjects:
            run_subjects = run_subjects if isinstance(run_subjects, list) else [run_subjects]
            run_subjects = [subj_id.removeprefix("sub-") for subj_id in map(str, run_subjects)]
            subj_ids = sorted([subj_id for subj_id in subj_ids if subj_id in run_subjects])

        # Setup extraction
        self._setup_extraction(layout, subj_ids, exclude_niftis, verbose)

        if self._n_cores:
            # Generate list of tuples for each subject
            args_list = [
                (
                    subj_id,
                    self._subject_info[subj_id]["prepped_files"],
                    self._subject_info[subj_id]["run_list"],
                    self._parcel_approach,
                    self._signal_clean_info,
                    self._task_info,
                    self._subject_info[subj_id]["tr"],
                    verbose,
                    flush,
                    parallel_log_config,
                )
                for subj_id in self._subject_ids
            ]

            parallel = Parallel(return_as="generator", n_jobs=self._n_cores, backend="loky")
            outputs = tqdm(
                parallel(delayed(_extract_timeseries)(*args) for args in args_list),
                desc="Processing Subjects",
                total=len(args_list),
                disable=not progress_bar,
            )

            for output in outputs:
                subject_timeseries, qc = output
                self._expand_dicts(subject_timeseries, qc)
        else:
            if parallel_log_config:
                LG.warning(
                    "`parallel_log_config` is only used for parallel processing. "
                    "The default logger can be modified by configuring either the root logger or a "
                    "logger for a specific module prior to package import."
                )

            for subj_id in tqdm(
                self._subject_ids, desc="Processing Subjects", disable=not progress_bar
            ):
                outputs = _extract_timeseries(
                    subj_id=subj_id,
                    **self._subject_info[subj_id],
                    parcel_approach=self._parcel_approach,
                    signal_clean_info=self._signal_clean_info,
                    task_info=self._task_info,
                    verbose=verbose,
                    flush=flush,
                )

                self._expand_dicts(*outputs)

        return self

    def _validate_get_bold_params(self) -> None:
        """Validates ``condition_tr_shift`` and ``slice_time_ref``."""
        if self._task_info["condition_tr_shift"] != 0:
            if (
                not isinstance(self._task_info["condition_tr_shift"], int)
                or self._task_info["condition_tr_shift"] < 0
            ):
                raise ValueError(
                    "`condition_tr_shift` must be a integer value equal to or greater than 0."
                )

            if self._task_info["condition"] is None:
                LG.warning("`condition_tr_shift` specified but `condition` is None.")

        if self._task_info["slice_time_ref"] != 0:
            if not isinstance(self._task_info["slice_time_ref"], (float, int)) or (
                self._task_info["slice_time_ref"] < 0 or (self._task_info["slice_time_ref"] > 1)
            ):
                raise ValueError("`slice_time_ref` must be a numerical value from 0 to 1.")

            if self._task_info["condition"] is None:
                LG.warning("`slice_time_ref` specified but `condition` is None.")

    @staticmethod
    @lru_cache(maxsize=4)
    def _call_layout(bids_dir: str, pipeline_name: Union[str]):
        """
        Returns ``BIDSLayout``.

        Raises
        ------
        ModuleNotFoundError
            The pybids package is not installed.

        Note
        ----
        The type hint for the returned (``BIDSLayout`)` is not added to allow certain functions in
        ``TimeseriesExtractor`` to be used on Windows machines that do not have pybids installed.
        """
        try:
            import bids
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "This function relies on the pybids package to query subject-specific files. "
                "If on Windows, pybids does not install by default to avoid long path error issues "
                "during installation. Try using `pip install pybids` or `pip install neurocaps[windows]`."
            )

        bids_dir = os.path.normpath(bids_dir).rstrip(os.path.sep)
        bids_dir = bids_dir.removesuffix("derivatives").rstrip(os.sep)

        if pipeline_name:
            pipeline_name = os.path.normpath(pipeline_name).lstrip(os.path.sep).rstrip(os.path.sep)
            pipeline_name = pipeline_name.removeprefix("derivatives").lstrip(os.path.sep)

            layout = bids.BIDSLayout(
                bids_dir, derivatives=os.path.join(bids_dir, "derivatives", pipeline_name)
            )
        else:
            layout = bids.BIDSLayout(bids_dir, derivatives=True)

        LG.info(f"{layout}")

        return layout

    def _setup_extraction(
        self, layout, subj_ids: list[str], exclude_niftis: list, verbose: bool
    ) -> None:
        """
        Get valid subjects (stored in ``self._subject_info``) and all required information needed
        to extract timeseries.

        Note
        ----
        The type hint for ``layout`` is not added to allow certain in ``TimeseriesExtractor`` to be
        used on Windows machines that do not have pybids installed.
        """
        base_dict = {"layout": layout, "subj_id": None}

        for subj_id in subj_ids:
            base_dict["subj_id"] = subj_id
            files = self._build_dict(base_dict)

            # Remove excluded file from the niftis list, which will prevent it from being processed
            if exclude_niftis and files["niftis"]:
                files["niftis"] = self._exclude(files["niftis"], exclude_niftis)

            # Get subject header
            subject_header = self._header(subj_id)

            # Check files
            skip, msg = self._check_files(files)
            if msg and verbose:
                LG.warning(subject_header + msg)
            if skip:
                continue

            # Ensure only a single session is present if session is None
            if not self._task_info["session"]:
                self._check_sess(files["niftis"], subject_header)

            # Generate a list of runs to iterate through based on runs in niftis
            check_runs = self._gen_runs(files["niftis"])
            if check_runs:
                run_list = self._filter_runs(check_runs, files)

                # Skip subject if no run has all needed files present
                if not run_list:
                    if verbose:
                        LG.warning(
                            f"{subject_header}"
                            "Timeseries Extraction Skipped: None of the necessary files "
                            "(i.e NifTIs, confound tsv files, confound json files, event files) "
                            "are from the same run."
                        )
                    continue

                if len(run_list) != len(check_runs):
                    if verbose:
                        LG.warning(
                            f"{subject_header}"
                            "Only the following runs available contain all required files: "
                            f"{', '.join(run_list)}."
                        )
            elif not check_runs and self._task_info["runs"]:
                if verbose:
                    requested_runs = [
                        f"run-{str(run)}" if "run-" not in str(run) else run
                        for run in self._task_info["runs"]
                    ]
                    LG.warning(
                        f"{subject_header}"
                        "Timeseries Extraction Skipped: Subject does not have any of the requested "
                        f"run IDs: {', '.join(requested_runs)}"
                    )
                    continue
            else:
                # Allows for nifti files that do not have the run- description
                run_list = [None]

            # Get repetition time for the subject
            tr = self._get_tr(files["bold_meta"], subject_header, verbose)

            # Add subject list to subject attribute. These are subjects that will be ran
            self._subject_ids.append(subj_id)

            # Store subject specific information
            self._subject_info[subj_id] = {"prepped_files": files, "tr": tr, "run_list": run_list}

    def _get_files(
        self,
        layout,
        extension: str,
        subj_id: str,
        scope: str = "derivatives",
        suffix: Union[str, None] = None,
        desc: Union[str, None] = None,
        event: bool = False,
        space: str = "attr",
    ):
        """
        Queries specific files (sorted lexicographically) using ``BidsLayout``.

        Note
        ----
        The type hint for ``layout`` is not added to allow certain in ``TimeseriesExtractor`` to be
        used on Windows machines that do not have pybids installed.
        """
        query_dict = {
            "scope": scope,
            "return_type": "file",
            "task": self._task_info["task"],
            "extension": extension,
            "subject": subj_id,
        }

        if desc:
            query_dict.update({"desc": desc})

        if suffix:
            query_dict.update({"suffix": suffix})

        if self._task_info["session"]:
            query_dict.update({"session": self._task_info["session"]})

        if not event and not desc:
            query_dict.update({"space": self._space if space == "attr" else space})

        return sorted(layout.get(**query_dict))

    def _build_dict(self, base: dict) -> dict:
        """Builds dictionary containing subject-specific files queried using ``BIDSLayout``."""
        files = {}
        files["niftis"] = self._get_files(**base, suffix="bold", extension="nii.gz")

        files["bold_meta"] = self._get_files(**base, suffix="bold", extension="json")
        if not files["bold_meta"]:
            files["bold_meta"] = self._get_files(
                **base, scope="raw", suffix="bold", extension="json", space=None
            )

        if self._task_info["condition"]:
            files["events"] = self._get_files(
                **base, scope="raw", suffix="events", extension="tsv", event=True
            )
        else:
            files["events"] = []

        files["confounds"] = self._get_files(**base, desc="confounds", extension="tsv")

        if self._signal_clean_info["n_acompcor_separate"]:
            files["confound_metas"] = self._get_files(**base, extension="json", desc="confounds")

        return files

    @staticmethod
    def _exclude(niftis: list, exclude_niftis: list) -> list:
        """Excludes certain NifTI files based on ``exclude_niftis``."""
        exclude_niftis = exclude_niftis if isinstance(exclude_niftis, list) else [exclude_niftis]
        return [nifti for nifti in niftis if os.path.basename(nifti) not in exclude_niftis]

    def _header(self, subj_id: str) -> str:
        """Creates base subject-specific header for logged messages."""
        sub_message = (
            f"[SUBJECT: {subj_id} | "
            f"SESSION: {self._task_info['session']} | "
            f"TASK: {self._task_info['task']}]"
        )
        subject_header = f"{sub_message} "

        return subject_header

    def _check_files(self, files: dict) -> tuple[Union[bool, None], Union[str, None]]:
        """
        Simple initial check to ensure the required files are needed based on certain
        parameters ``__init__``.
        """
        skip, msg = None, None

        if not files["niftis"]:
            skip = True
            msg = (
                "Timeseries Extraction Skipped: No NifTI files were found or all NifTI files "
                "were excluded."
            )

        if self._signal_clean_info["use_confounds"]:
            if not files["confounds"]:
                skip = True
                msg = (
                    "Timeseries Extraction Skipped: `use_confounds` is requested but no "
                    "confound files found."
                )

            if self._signal_clean_info["n_acompcor_separate"] and not files.get("confound_metas"):
                skip = True
                msg = (
                    "Timeseries Extraction Skipped: No confound metadata file found, which is "
                    "needed to locate the first n components of the white-matter and cerebrospinal "
                    "fluid masks separately."
                )

            if self._task_info["condition"] and not files["events"]:
                skip = True
                msg = (
                    "Timeseries Extraction Skipped: `condition` is specified but no event files "
                    "found."
                )

        return skip, msg

    def _check_sess(self, niftis: list, subject_header: str) -> None:
        """
        Checks if all subject's NifTI's are from a single session and returns an error if
        different sessions are detected.
        """
        ses_list = []

        for nifti in niftis:
            if "ses-" in os.path.basename(nifti):
                ses_list.append(re.search(r"ses-(\S+?)_", os.path.basename(nifti))[0][:-1])

        ses_list = sorted(set(ses_list))
        if len(ses_list) > 1:
            raise ValueError(
                f"{subject_header}"
                "`session` not specified but subject has more than one session: "
                f"{', '.join(ses_list)}. In order to continue timeseries extraction, the "
                "specific session to extract must be specified using `session`."
            )

    def _gen_runs(self, niftis: list) -> set:
        """
        Gets all the runs for a specific subject and filters if specific runs are requested.
        Returns set of run IDs sorted lexicographically.
        """
        check_runs = []

        for nifti in niftis:
            if "run-" in os.path.basename(niftis[0]):
                check_runs.append(re.search(r"run-(\S+?)_", os.path.basename(nifti))[0][:-1])

        check_runs = set(check_runs)

        requested_runs = {}
        if self._task_info["runs"]:
            requested_runs = {f"run-{run}" for run in self._task_info["runs"]}

        return (
            sorted(check_runs.intersection(requested_runs))
            if requested_runs
            else sorted(check_runs)
        )

    def _filter_runs(self, check_runs: list, files: dict) -> list:
        """Filters runs by checking if all required files have the same run ID."""
        run_list = []

        # Check if at least one run has all files present
        for run in check_runs:
            bool_list = []

            # Assess is any of these returns True
            bool_list.append(any(f"{run}_" in file for file in files["niftis"]))

            if self._task_info["condition"]:
                bool_list.append(any(f"{run}_" in file for file in files["events"]))

            if self._signal_clean_info["use_confounds"]:
                bool_list.append(any(f"{run}_" in file for file in files["confounds"]))

                if self._signal_clean_info["n_acompcor_separate"]:
                    bool_list.append(any(f"{run}_" in file for file in files["confound_metas"]))

            # Append runs that contain all needed files
            if all(bool_list):
                run_list.append(run)

        return run_list

    def _get_tr(
        self, bold_meta: str, subject_header: str, verbose: bool
    ) -> Union[float, int, None]:
        """Gets repetition time."""
        try:
            if self._task_info["tr"]:
                tr = self._task_info["tr"]
            else:
                with open(bold_meta[0], "r") as json_file:
                    tr = json.load(json_file)["RepetitionTime"]
        except (IndexError, KeyError, json.JSONDecodeError) as e:
            base_msg = (
                "`tr` not specified and could not be extracted since using the first BOLD "
                "metadata file"
            )
            base_msg += " due to " + (
                f"there being no BOLD metadata files for [TASK: {self._task_info['task']}]"
                if str(type(e).__name__) == "IndexError"
                else f"{type(e).__name__} - {str(e)}"
            )

            if self._task_info["condition"]:
                raise ValueError(
                    f"{subject_header}"
                    f"{base_msg}" + " The `tr` must be provided when `condition` is specified."
                )
            elif any(
                [
                    self._signal_clean_info["masker_init"]["high_pass"],
                    self._signal_clean_info["masker_init"]["low_pass"],
                ]
            ):
                raise ValueError(
                    f"{subject_header}"
                    f"{base_msg}"
                    + " The `tr` must be provided when `high_pass` or `low_pass` is specified."
                )
            elif self._signal_clean_info["fd_threshold"].get("interpolate"):
                raise ValueError(
                    "`tr` must be provided when interpolation of censored volumes is required."
                )
            else:
                if verbose:
                    LG.warning(
                        f"{subject_header}"
                        f"{base_msg}" + " `tr` has been set to None but extraction will continue."
                    )
                    tr = None

        return tr

    def _expand_dicts(
        self, subject_timeseries: Union[SubjectTimeseries, None], qc: Union[dict, None]
    ) -> None:
        """
        Aggregates individual subject timeseries and qc dictionaries into ``self._subject_timeseries``
        and ``self._qc``.
        """
        if isinstance(subject_timeseries, dict):
            self._subject_timeseries.update(subject_timeseries)
            if qc:
                self._qc.update(qc)

    @staticmethod
    def _raise_error(attr_name: str, msg: str) -> None:
        """Raises error if ``_subject_timeseries`` pr ``self._qc`` is not available."""
        if attr_name == "_subject_timeseries":
            raise AttributeError(
                f"{msg} since `self.subject_timeseries` is None, either run `self.get_bold()` or "
                "assign a valid timeseries dictionary to `self.subject_timeseries`."
            )
        else:
            raise AttributeError(f"{msg} since `self.qc` is None, run `self.get_bold()` first.")

    def timeseries_to_pickle(self, output_dir: str, filename: Optional[str] = None) -> Self:
        """
        Save the Extracted Subject Timeseries.

        Saves the extracted timeseries stored in the ``self.subject_timeseries`` dictionary as a
        pickle file (ending in ".pkl") with Joblib.

        Parameters
        ----------
        output_dir: :obj:`str`
            Directory to save a pickle file to. The directory will be created if it does not exist.

        filename: :obj:`str` or :obj:`None`, default=None
            Name of the file. If None, will use "subject_timeseries.pkl" as default.

        Returns
        -------
        self
        """
        save_filename = self._create_output_filename(
            output_dir, filename, attr_name="_subject_timeseries", call="timeseries_to_pickle"
        )

        io_utils._serialize(self._subject_timeseries, output_dir, save_filename, use_joblib=True)

        return self

    def report_qc(
        self,
        output_dir: Optional[str] = None,
        filename: Optional[str] = None,
        return_df: bool = True,
    ) -> Union[DataFrame, None]:
        """
        Report Quality Control Information.

        Converts per-subject, per-run quality control metrics (related to framewise displacement and
        dummy volumes) stored in ``self.qc`` dictionary to a pandas DataFrame for analysis and
        reporting.

        .. note:: In versions >= 0.27.0, "NaN" is used to denote a value is not applicable.

        Parameters
        ----------
        output_dir: :obj:`str`, default=None
            Directory to save a pickle file to. The directory will be created if it does not exist.
            The file will not be saved if an output directory is not provided.

        filename: :obj:`str` or :obj:`None`, default=None
            Name of the file with or without the "csv" extension.

        return_df: :obj:`bool`, default=True
            If True, returns the dataframe.

        Returns
        -------
        pandas.Dataframe
            Pandas dataframe containing the colums: "Subject_ID", "Run", "Frames_Scrubbed",
            "Frames_Interpolated", "Mean_High_Motion_Length", "Std_High_Motion_Length", "Mean_FD",
            "Std_FD", and "N_Dummy_Scans".

        Important
        ---------
        **Reporting**: Statistics for each subject's run are specific to the specified
        ``fd_threshold`` and ``dummy_scans``. Additionally, all reported statistics for framewise
        displacement are specific to steady-state volumes (if ``dummy_scans`` is not None), as they
        are only computed after accounting for dummy volumes.

        **Note:** If ``condition`` was specified in ``self.get_bold()``, all the reported statistics
        related to framewise displacement are specific to the frames assigned to the condition (as
        determined by the equation in "Extraction of Task Conditions" in the documentation of
        ``self.get_bold()``), which are treated as one continuous event window for computational
        simplicity. The number of reported dummy volumes is not adjusted for the condition.

        **Mean & Standard Deviation of Framewise Displacement:** "Mean_FD" and "Std_FD" represent
        the statistics prior to scrubbing or interpolating.

        **Censored & Interpolated Frames**: The metrics for scrubbed frames and interpolated frames
        are independent. For instance, 2 scrubbed frames and 3 interpolated frames means that 3
        frames were interpolated while 2 were scrubbed due to excessive motion.

        **Note:** Only censored frames with bounded by non-censored frames on both sides are
        interpolated, while censored frames at the edge of the timeseries (including frames that
        border censored edges) are always scrubbed and counted in "Frames_Scrubbed". For instance,
        if the full sample mask is computed as ``[0, 0, 1, 0, 0, 1, 0]`` where "0" are censored and
        "1" is not censored, then when no interpolation is requested, then "Frames_Scrubbed" will be
        5 and "Frames_Interpolated" will be 0. If interpolation is requested, then "Frames_Scrubbed"
        would be 3 and "Frames_Interpolated" would be 2 (indexes 3 and 4).

        **High Motion Length Computation**: "Mean_High_Motion_Length" and "Std_High_Motion_Length"
        represent the average length and population standard deviation of consecutive frames flagged
        for high-motion frames, respectively.
        """
        save_filename = self._create_output_filename(
            output_dir, filename, attr_name="_qc", call="report_qc"
        )
        assert self._qc, "No quality control information to report."

        # Build df
        df = DataFrame(
            columns=[
                "Subject_ID",
                "Run",
                "Mean_FD",
                "Std_FD",
                "Frames_Scrubbed",
                "Frames_Interpolated",
                "Mean_High_Motion_Length",
                "Std_High_Motion_Length",
                "N_Dummy_Scans",
            ]
        )
        for subject in self._qc:
            for run in self._qc[subject]:
                df.loc[len(df)] = [
                    subject,
                    run,
                    self._qc[subject][run]["mean_fd"],
                    self._qc[subject][run]["std_fd"],
                    self._qc[subject][run]["frames_scrubbed"],
                    self._qc[subject][run]["frames_interpolated"],
                    self._qc[subject][run]["mean_high_motion_length"],
                    self._qc[subject][run]["std_high_motion_length"],
                    self._qc[subject][run]["n_dummy_scans"],
                ]

        if output_dir:
            df.to_csv(os.path.join(output_dir, save_filename), sep=",", index=0)

        if return_df:
            return df

    def _create_output_filename(
        self, output_dir: Union[str, None], filename: Union[str, None], attr_name: str, call: str
    ) -> Union[str, None]:
        """Checks if the required attribute is present then creates the output filename."""
        if not hasattr(self, attr_name):
            self._raise_error(
                attr_name,
                msg=(
                    "Cannot save pickle file"
                    if attr_name == "_subject_timeseries"
                    else "Cannot save csv file"
                ),
            )

        if not output_dir:
            return None
        else:
            io_utils._makedir(output_dir)

        ext = "pkl" if call == "timeseries_to_pickle" else "csv"

        if filename is None:
            save_filename = (
                "subject_timeseries.pkl" if call == "timeseries_to_pickle" else "report_qc.csv"
            )
        else:
            save_filename = f"{os.path.splitext(filename.rstrip())[0].rstrip()}.{ext}"

        return save_filename

    def visualize_bold(
        self,
        subj_id: Union[int, str],
        run: Optional[Union[int, str]] = None,
        roi_indx: Optional[Union[int, str, list[str], list[int]]] = None,
        region: Optional[str] = None,
        show_figs: bool = True,
        output_dir: Optional[str] = None,
        filename: Optional[str] = None,
        as_pickle: bool = False,
        **kwargs,
    ) -> Self:
        """
        Plot the Extracted Subject Timeseries.

        Visualize the extracted BOLD timeseries data of nodes (Region-of-Interests [ROIs]) or
        regions (anatomical regions/networks) for a specific subject and run.

        Parameters
        ----------
        subj_id: :obj:`str` or :obj:`int`
            The ID of the subject.

        run: :obj:`int`, :obj:`str`, or :obj:`None`, default=None
            The run ID of the subject to plot. Must be specified if multiple runs exist for a given
            subject.

        roi_indx: :obj:`int`, :obj:`str`, :obj:`list[int]`, :obj:`list[int]` or :obj:`None`, default=None
            The indices of the parcellation nodes to plot. See "nodes" in ``self.parcel_approach``
            for valid nodes.

        region: :obj:`str` or :obj:`None`, default=None
            The region of the parcellation to plot. If not None, all nodes in the specified region
            will be averaged then plotted. See "regions" in ``self.parcel_approach`` for valid region.

        show_figs: :obj:`bool`, default=True
            Display figures.

        output_dir: :obj:`str` or :obj:`None`, default=None
            Directory to save plot as png image. The directory will be created if it does not exist.
            If None, plot will not be saved.

        filename: :obj:`str` or :obj:`None`, default=None
            Name of the file without the extension.

        as_pickle: :obj:`bool`, default=False
            When ``output_dir`` is specified, plots are saved as pickle file, which can be further
            modified, instead of png images.

            .. versionadded:: 0.26.5

        **kwargs:
            Keyword arguments used when saving figures. Valid keywords include:

            - dpi: :obj:`int`, default=300 -- Dots per inch for the figure.
            - figsize: :obj:`tuple`, default=(11, 5) -- Size of the figure in inches.
            - bbox_inches: :obj:`str` or :obj:`None`, default="tight" -- Alters size of the
              whitespace in the saved image.

        Returns
        -------
        self

        Note
        ----
        **Parcellation Approach**: the "nodes" and "regions" subkeys are required in
        ``parcel_approach``.
        """

        if not hasattr(self, "_subject_timeseries"):
            self._raise_error(attr_name="_subject_timeseries", msg="Cannot plot bold data")

        subj_id = str(subj_id).removeprefix("sub-")
        if subj_id not in self._subject_timeseries:
            raise KeyError(f"Subject {subj_id} is not available in `self._subject_timeseries`.")

        if len(subject_runs := list(self._subject_timeseries[subj_id])) > 1 and not run:
            raise ValueError(
                "`run` must be specified when multiple runs exist. "
                f"Runs available for sub-{subj_id}: {', '.join(subject_runs)}."
            )

        if roi_indx is None and region is None:
            raise ValueError("either `roi_indx` or `region` must be specified.")

        if roi_indx is not None and region is not None:
            raise ValueError("`roi_indx` and `region` can not be used simultaneously.")

        io_utils._issue_file_warning("filename", filename, output_dir)

        # Defaults
        plot_dict = _check_kwargs(_PlotDefaults.visualize_bold(), **kwargs)

        # Obtain the column indices associated with the rois
        parcellation_name = list(self._parcel_approach)[0]
        if roi_indx is not None:
            plot_indxs = self._get_roi_indices(roi_indx, parcellation_name)
        else:
            plot_indxs = self._get_region_indices(region, parcellation_name)

        plt.figure(figsize=plot_dict["figsize"])

        run = None if not run else str(run)
        run = subject_runs[0] if not run else f"run-{run.removeprefix('run-')}"
        timeseries = self._subject_timeseries[subj_id][run]
        if roi_indx or roi_indx == 0:
            plt.plot(range(1, timeseries.shape[0] + 1), timeseries[:, plot_indxs])

            if isinstance(roi_indx, (int, str)) or (
                isinstance(roi_indx, list) and len(roi_indx) == 1
            ):
                if isinstance(roi_indx, int):
                    roi_title = self._parcel_approach[parcellation_name]["nodes"][roi_indx]
                elif isinstance(roi_indx, str):
                    roi_title = roi_indx
                else:
                    roi_title = roi_indx[0]
                plt.title(roi_title)
        else:
            plt.plot(range(1, timeseries.shape[0] + 1), np.mean(timeseries[:, plot_indxs], axis=1))
            plt.title(region)

        plt.xlabel("TR")

        if output_dir:
            io_utils._makedir(output_dir)

            if filename:
                save_filename = f"{os.path.splitext(filename.rstrip())[0].rstrip()}.png"
            else:
                save_filename = f"subject-{subj_id}_{run}_timeseries.png"

            _PlotFuncs.save_fig(plt.gcf(), output_dir, save_filename, plot_dict, as_pickle)

        _PlotFuncs.show(show_figs)

        return self

    def _get_roi_indices(self, roi_indx: Union[int, str, list], parcellation_name: str) -> NDArray:
        """Gets the indices for a specified node or nodes from ``self._parcel_approach``."""
        if isinstance(roi_indx, int):
            plot_indxs = roi_indx
        elif isinstance(roi_indx, str):
            # Check if parcellation_approach is custom
            if "Custom" in self._parcel_approach and "nodes" not in self._parcel_approach["Custom"]:
                _check_parcel_approach(parcel_approach=self._parcel_approach, call="visualize_bold")

            plot_indxs = list(self._parcel_approach[parcellation_name]["nodes"]).index(roi_indx)
        else:
            if all(isinstance(indx, int) for indx in roi_indx):
                plot_indxs = np.array(roi_indx)
            elif all(isinstance(indx, str) for indx in roi_indx):
                # Check if parcellation_approach is custom
                if (
                    "Custom" in self._parcel_approach
                    and "nodes" not in self._parcel_approach["Custom"]
                ):
                    _check_parcel_approach(
                        parcel_approach=self._parcel_approach, call="visualize_bold"
                    )

                plot_indxs = np.array(
                    [
                        list(self._parcel_approach[parcellation_name]["nodes"]).index(index)
                        for index in roi_indx
                    ]
                )
            else:
                raise ValueError(
                    "All elements in `roi_indx` need to be all strings or all integers."
                )

        return plot_indxs

    def _get_region_indices(self, region: str, parcellation_name: str) -> NDArray:
        """Gets the indices for a specified region from ``self._parcel_approach``."""
        if "Custom" in self._parcel_approach:
            if "regions" not in self._parcel_approach["Custom"]:
                _check_parcel_approach(parcel_approach=self._parcel_approach, call="visualize_bold")
            else:
                plot_indxs = np.array(
                    list(self._parcel_approach["Custom"]["regions"][region]["lh"])
                    + list(self._parcel_approach["Custom"]["regions"][region]["rh"])
                )
        else:
            plot_indxs = np.array(
                [
                    index
                    for index, label in enumerate(self._parcel_approach[parcellation_name]["nodes"])
                    if region in label
                ]
            )

        return plot_indxs



"""Internal function for performing cluster selection with or without joblib"""

from sklearn.cluster import KMeans
from sklearn.metrics import davies_bouldin_score, calinski_harabasz_score, silhouette_score


def _run_kmeans(n_cluster, configs, concatenated_timeseries, method):
    """
    Uses scikit-learn to perform k-means clustering on concatenated timeseries data in both
    sequential and parallel contexts. Also uses scikit-learn to provide cluster performance metrics.
    """
    model = KMeans(n_clusters=n_cluster, **configs, verbose=0).fit(concatenated_timeseries)

    # Only return model when no cluster selection chosen
    if method is None:
        return model

    cluster_labels = model.labels_
    if method == "davies_bouldin":
        performance = {n_cluster: davies_bouldin_score(concatenated_timeseries, cluster_labels)}
    elif method == "elbow":
        performance = {n_cluster: model.inertia_}
    elif method == "silhouette":
        performance = {
            n_cluster: silhouette_score(concatenated_timeseries, cluster_labels, metric="euclidean")
        }
    else:
        # Variance Ratio
        performance = {n_cluster: calinski_harabasz_score(concatenated_timeseries, cluster_labels)}

    model_dict = {n_cluster: model}

    return performance, model_dict



"""Function for merging timeseries data across dictionaries."""

from typing import Union, Optional

import numpy as np

import neurocaps._utils.io as io_utils
from ..typing import SubjectTimeseries


def merge_dicts(
    subject_timeseries_list: Union[list[SubjectTimeseries], list[str]],
    return_merged_dict: bool = True,
    return_reduced_dicts: bool = False,
    output_dir: Optional[Union[str, str]] = None,
    filenames: Optional[list[str]] = None,
    save_reduced_dicts: bool = False,
) -> Union[dict[str, SubjectTimeseries], None]:
    """
    Merge Participant Timeseries Across Multiple Sessions or Tasks.

    Merge subject timeseries data across dictionaries, concatenating matching run IDs. Only subjects
    present across all input dictionaries are included in the merged output.

    For example, if three dictionaries are provided contain subject 1 with:

        - dict 1: run-1 (resting-state)
        - dict 2: run-1 and run-2 (stroop)
        - dict 3: run-3 (n-back)

    Then subject 1 in the final merged dictionary will contain:

        - run-1: concatenated timeseries from dict 1 and dict 2 (resting-state + stroop)
        - run-2: timeseries from dict 2 (stroop)
        - run-3: timeseries from dict 3 (n-back)

    This function supports workflows for identifying similar CAPs across tasks or sessions.
    Specifically, using the merged dictionary as input for ``CAP.get_caps`` and the filtered input
    dictionaries, containing only subjects present in the merged dictionary, as inputs for
    ``CAP.calculate_metrics`` to compute participant-wise CAP metrics for each task.

    Parameters
    ----------
    subject_timeseries_list: :obj:`list[SubjectTimeseries]` or :obj:`list[str]`
        A list where each element consist of a dictionary mapping subject IDs to their run IDs and
        associated timeseries (TRs x ROIs) as a NumPy array. Can also be a list consisting of paths
        to pickle files containing this same structure. Refer to documentation for
        ``SubjectTimeseries`` in the "See Also" section for an example structure.

    return_merged_dict: :obj:`bool`, default=True
        If True, returns a single dictionary containing the merged dictionary under a key named
        "merged".

    return_reduced_dicts: :obj:`bool`, default=False
        If True, returns a single dictionary containing the input dictionaries filtered to only
        include subjects present in the merged dictionary. Keys are named "dict_{0}" where {0}
        corresponds to the dictionary's position in the input list.

    output_dir: :obj:`str` or :obj:`None`, default=None
        Directory to save the merged or reduced dictionaries as pickle files. The directory will be
        created if it does not exist. For the reduced dictionaries to be saved, ``save_reduced_dicts``
        must be set to True. If ``save_reduced_dicts`` is False and ``output_dir`` is provided, only
        the merged dictionary will be saved. Dictionaries will not be saved if None.

    filenames: :obj:`list[str]` or :obj:`None`, default=None
        A list of file names for saving dictionaries when ``output_dir`` is provided.

        If ``save_reduced_dicts`` is False:

            - Provide a single name, which will be used to save the merged dictionary.

        If ``save_reduced_dicts`` is True:

            - Provide N+1 names (where N is the length of ``subject_timeseries_list``): N names for
            individual reduced dictionaries followed by one name for the merged dictionary. Names
            are assigned by input position order.

        *Note*: Full paths are handled using basename and extensions are ignored. If None, uses
        default names - "subject_timeseries_{0}_reduced.pkl" (where {0} indicates the original input
        order) and "merged_subject_timeseries.pkl" for the merged dictionary.

    save_reduced_dicts: :obj:`bool` or None, default=False
        If True and the ``output_dir`` is provided, then the reduced dictionaries are saved as
        pickle files.

    Returns
    -------
    dict[str, SubjectTimeseries]
        A nested dictionary containing the merged subject timeseries and reduced subject timeseries
        if either ``return_merged_dict`` or ``return_reduced_dict`` are True.

    See Also
    --------
    :data:`neurocaps.typing.SubjectTimeseries`
        Type definition for the subject timeseries dictionary structure.

    References
    ----------
    Kupis, L., Romero, C., Dirks, B., Hoang, S., Parladé, M. V., Beaumont, A. L., Cardona, S. M.,
    Alessandri, M., Chang, C., Nomi, J. S., & Uddin, L. Q. (2020). Evoked and intrinsic brain
    network dynamics in children with autism spectrum disorder. NeuroImage: Clinical, 28, 102396.
    https://doi.org/10.1016/j.nicl.2020.102396
    """

    assert isinstance(subject_timeseries_list, list), "`subject_timeseries_list` must be a list."
    assert (
        len(subject_timeseries_list) > 1
    ), "Merging cannot be done with less than two dictionaries or files."

    io_utils._issue_file_warning("filenames", filenames, output_dir)

    # Only perform IO operation once
    new_timeseries_list = []
    for curr_dict in subject_timeseries_list:
        curr_dict = io_utils._get_obj(curr_dict, needs_deepcopy=False)
        new_timeseries_list.append(curr_dict)

    # Get common subject ids
    subject_set = {}
    for curr_dict in new_timeseries_list:
        if not subject_set:
            subject_set = set(curr_dict)

        subject_set = subject_set.intersection(list(curr_dict))

    # Order subjects
    intersect_subjects = sorted(list(subject_set))

    # Collect timeseries array in a temporary dictionary as lists to reduce repeated np.vstack calls
    temp_dict = {}
    for curr_dict in new_timeseries_list:
        for subj_id in intersect_subjects:
            if subj_id not in temp_dict:
                temp_dict[subj_id] = {}

            for curr_run, arr in curr_dict[subj_id].items():
                if curr_run not in temp_dict[subj_id]:
                    temp_dict[subj_id][curr_run] = [arr]
                else:
                    temp_dict[subj_id][curr_run].append(arr)

    # Now use temporary dict to create the final merged dictionary
    subject_timeseries_merged = {}
    for subj_id, runs_dict in temp_dict.items():
        subject_timeseries_merged[subj_id] = {}
        for curr_run, arr_list in runs_dict.items():
            # Ensure memory address of numpy array is different when only a single array
            if len(arr_list) == 1:
                subject_timeseries_merged[subj_id][curr_run] = arr_list[0].copy()
            else:
                subject_timeseries_merged[subj_id][curr_run] = np.vstack(arr_list)

        # Sort runs lexicographically
        if list(subject_timeseries_merged[subj_id]) != sorted(
            subject_timeseries_merged[subj_id].keys()
        ):
            subject_timeseries_merged[subj_id] = {
                run_id: subject_timeseries_merged[subj_id][run_id]
                for run_id in sorted(subject_timeseries_merged[subj_id].keys())
            }

    del temp_dict

    modified_dicts = {}
    # Obtain the reduced dictionaries
    if return_reduced_dicts or (save_reduced_dicts and output_dir):
        for indx, curr_dict in enumerate(new_timeseries_list):
            modified_dicts[f"dict_{indx}"] = {}

            for subj_id in subject_timeseries_merged:
                # Diff memory address from original array
                modified_dicts[f"dict_{indx}"].update({subj_id: curr_dict[subj_id].copy()})

    if return_merged_dict or output_dir:
        modified_dicts["merged"] = subject_timeseries_merged

    if output_dir:
        message = (
            "Length of `file_names` must be equal to 1 if `save_reduced_dicts`is False or the "
            "length of `subject_timeseries_list` + 1 if `save_reduced_dicts` is True."
        )

        # Convert to list if string
        if isinstance(filenames, str):
            filenames = [filenames]

        io_utils._dicts_to_pickles(
            output_dir=output_dir,
            dict_list=modified_dicts,
            call="merge",
            filenames=filenames,
            message=message,
            save_reduced_dicts=save_reduced_dicts,
        )

    if return_merged_dict or return_reduced_dicts:
        return modified_dicts



"""
Internal IO functions for creating directories, issuing warnings, serializing (with
Joblib or pickle), unserializing (with Joblib), and additional utilities.
"""

import os, copy, pickle
from typing import Any, Union

import joblib

from .logger import _logger
from ..exceptions import UnsupportedFileExtensionError

LG = _logger(__name__)


def _makedir(output_dir: str) -> None:
    """Creates non-existent directory."""
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)


def _issue_file_warning(param_name: str, param: str, output_dir: str) -> None:
    """Issues warning."""
    if param is not None and not output_dir:
        LG.warning(
            f"`{param_name}` supplied but no `output_dir` specified. Files will not be saved."
        )


def _filename(
    base_name: str, add_name: Union[str, None], pos: str, ext: Union[None, str] = None
) -> str:
    """
    Adds the file extension and the prefix or suffix to the file name depending on if ``pos``
    is "prefix" or "suffix".
    """
    if not add_name:
        return f"{base_name}.{ext}" if ext else base_name

    if pos == "prefix":
        # Use - since prefix is currently for ``CAP.calculate_metrics``
        add_name = os.path.splitext(add_name.rstrip())[0].rstrip()
        filename = f"{add_name}-{base_name}"
    else:
        filename = f"{base_name}_{add_name.rstrip().replace(' ', '_')}"

    filename = filename.replace(" ", "_")

    return f"{filename}.{ext}" if ext else filename


def _serialize(obj: Any, output_dir: str, save_filename: str, use_joblib=False) -> None:
    """
    Serialize an object with standard pickle or Joblib. Internally, plots use pickle for
    compatibility with pickle or Joblib.
    """
    with open(os.path.join(output_dir, save_filename), "wb") as f:
        joblib.dump(obj, f) if use_joblib else pickle.dump(obj, f)


def _get_obj(obj: Any, needs_deepcopy=True) -> Any:
    """Determines if object needs deepcopy or needs to be unserialized."""
    if not isinstance(obj, str):
        obj = copy.deepcopy(obj) if needs_deepcopy else obj
        return obj
    else:
        return _unserialize(obj)


def _unserialize(file: str) -> Any:
    """Opens serialized objects such as ``subject_timeseries`` or ``parcel_approach``."""
    base_ext = (".pkl", ".pickle", ".joblib")
    supported_ext = (*base_ext, *[f"{x}.gz" for x in base_ext])
    if file.endswith(supported_ext):
        with open(file, "rb") as f:
            obj = joblib.load(f)
    else:
        raise UnsupportedFileExtensionError(
            "Serialized files must end with one of the following extensions: "
            "'.pkl', '.pickle', '.joblib'."
        )

    return obj


def _dicts_to_pickles(
    output_dir: str,
    dict_list: list[dict],
    call: str,
    filenames: Union[str, None] = None,
    message: Union[str, None] = None,
    save_reduced_dicts: bool = False,
) -> None:
    """
    Saves dictionaries containing NumPy arrays as pickles for the ``merge_dicts``,
    ``standardize``, and ``change_dtypes`` functions.
    """
    if not filenames:
        saved_filenames = _create_default_filenames(dict_list, call, save_reduced_dicts)
    else:
        saved_filenames = [
            f"{os.path.splitext(os.path.basename(name).rstrip())[0].rstrip()}.pkl"
            for name in filenames
        ]

    if message is None:
        message = (
            "Length of `filenames` list  must be the same length as `subject_timeseries_list`."
        )

    if call == "merge" and save_reduced_dicts is False:
        dict_list = {"merged": dict_list["merged"]}

    if filenames:
        assert len(filenames) == len(list(dict_list)), message

    _makedir(output_dir)

    for i, dict_name in enumerate(list(dict_list)):
        with open(os.path.join(output_dir, saved_filenames[i]), "wb") as f:
            joblib.dump(dict_list[dict_name], f)


def _create_default_filenames(
    dict_list: list[dict], call: str, save_reduced_dicts: bool
) -> list[str]:
    """
    Generates default names for dictionaries created by the  ``merge_dicts``, ``standardize``, and
    ``change_dtypes`` functions.
    """
    if call == "merge":
        if save_reduced_dicts:
            default_filenames = [
                (
                    f"subject_timeseries_{key.split('_')[-1]}_reduced.pkl"
                    if "dict" in key
                    else "merged_subject_timeseries.pkl"
                )
                for key in list(dict_list)
            ]
        else:
            default_filenames = ["merged_subject_timeseries.pkl"]
    else:
        if call == "standardize":
            suffix = "standardized"
        else:
            dict_name = list(dict_list)[0]
            first_sub = list(dict_list[dict_name])[0]
            run_name = list(dict_list[dict_name][first_sub])[0]
            suffix = f"dtype-{str(dict_list[dict_name][first_sub][run_name].dtype)}"

        default_filenames = [
            f"subject_timeseries_{key.split('_')[-1]}_{suffix}.pkl" for key in list(dict_list)
        ]

    return default_filenames



"""
Internal logging function and class for flushing. Logger configured to respect user logging
configuration prior to package import while defaulting to console logging if no configurations are
provided.
"""

import logging, sys
from logging.handlers import QueueHandler

# Global variables to determine if a handler is user defined or defined by OS
_USER_ROOT_HANDLER = None
_USER_MODULE_HANDLERS = {}


class _Flush(logging.StreamHandler):
    def emit(self, record: logging.LogRecord) -> None:
        super().emit(record)
        self.flush()


def _logger(name, flush=False, top_level=True, parallel_log_config=None):
    global _USER_ROOT_HANDLER, _USER_MODULE_HANDLERS

    logger = logging.getLogger(name)
    parallel_module = "neurocaps._utils.extraction.extract_timeseries"

    if top_level is True:
        _USER_ROOT_HANDLER = logging.getLogger().hasHandlers()
        _USER_MODULE_HANDLERS[logger.name] = logger.handlers

    # Special case for parallel config to pass a user-defined logger specifically for parallel
    # processing.
    if logger.name == parallel_module and not top_level and parallel_log_config:
        # Only QueueHandler will be in handler list
        logger.handlers.clear()
        if "queue" in parallel_log_config:
            logger.addHandler(QueueHandler(parallel_log_config["queue"]))
        if "level" in parallel_log_config:
            logger.setLevel(parallel_log_config["level"])

    if not logger.level:
        logger.setLevel(logging.INFO)

    # Check if user defined root handler or assigned a specific handler for module
    default_handlers = _USER_ROOT_HANDLER or _USER_MODULE_HANDLERS[logger.name]

    # Propagate to root if no user-defined module or a root handler is detected
    logger.propagate = (
        False if _USER_MODULE_HANDLERS[logger.name] or not _USER_ROOT_HANDLER else True
    )

    # Add or messages will repeat several times due to multiple handlers if same name used
    if not default_handlers and not (
        parallel_log_config or (logger.name == parallel_module and top_level)
    ):
        # Safeguard; ensure a clean state for "extract_timeseries" since it is used in parallel and
        # sequential contexts
        if logger.name == parallel_module:
            logger.handlers.clear()

        if flush:
            handler = _Flush(sys.stdout)
        else:
            handler = logging.StreamHandler(sys.stdout)

        handler.setFormatter(logging.Formatter("%(asctime)s %(name)s [%(levelname)s] %(message)s"))
        logger.addHandler(handler)

    return logger



"""Classes to centralize plotting defaults and functions."""

import inspect, os
from typing import Union

import matplotlib.pyplot as plt, seaborn
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from pandas import DataFrame

import neurocaps._utils.io as io_utils


class _PlotDefaults:
    """Default plotting options."""

    @staticmethod
    def visualize_bold() -> dict:
        return {"dpi": 300, "figsize": (11, 5), "bbox_inches": "tight"}

    @staticmethod
    def get_caps() -> dict:
        return {"dpi": 300, "figsize": (8, 6), "step": None, "bbox_inches": "tight"}

    @staticmethod
    def caps2plot() -> dict:
        return {
            "dpi": 300,
            "figsize": (8, 6),
            "fontsize": 14,
            "hspace": 0.2,
            "wspace": 0.2,
            "xticklabels_size": 8,
            "yticklabels_size": 8,
            "cbarlabels_size": 8,
            "shrink": 0.8,
            "nrow": None,
            "ncol": None,
            "suptitle_fontsize": 20,
            "tight_layout": True,
            "rect": [0, 0.03, 1, 0.95],
            "sharey": True,
            "xlabel_rotation": 0,
            "ylabel_rotation": 0,
            "annot": False,
            "annot_kws": None,
            "fmt": ".2g",
            "linewidths": 0,
            "linecolor": "black",
            "cmap": "coolwarm",
            "edgecolors": None,
            "alpha": None,
            "hemisphere_labels": False,
            "borderwidths": 0,
            "vmin": None,
            "vmax": None,
            "bbox_inches": "tight",
        }

    @staticmethod
    def caps2corr() -> dict:
        return {
            "dpi": 300,
            "figsize": (8, 6),
            "fontsize": 14,
            "xticklabels_size": 8,
            "yticklabels_size": 8,
            "shrink": 0.8,
            "cbarlabels_size": 8,
            "xlabel_rotation": 0,
            "ylabel_rotation": 0,
            "annot": False,
            "linewidths": 0,
            "linecolor": "black",
            "cmap": "coolwarm",
            "fmt": ".2g",
            "borderwidths": 0,
            "edgecolors": None,
            "alpha": None,
            "bbox_inches": "tight",
            "annot_kws": None,
            "vmin": None,
            "vmax": None,
        }

    @staticmethod
    def caps2surf() -> dict:
        return {
            "dpi": 300,
            "title_pad": -3,
            "cmap": "cold_hot",
            "cbar_kws": {"location": "bottom", "n_ticks": 3},
            "size": (500, 400),
            "layout": "grid",
            "zoom": 1.5,
            "views": ["lateral", "medial"],
            "alpha": 1,
            "zero_transparent": True,
            "as_outline": False,
            "brightness": 0.5,
            "figsize": None,
            "scale": (2, 2),
            "surface": "inflated",
            "color_range": None,
            "bbox_inches": "tight",
            "outline_alpha": 1,
        }

    @staticmethod
    def caps2radar() -> dict:
        return {
            "scale": 2,
            "height": 800,
            "width": 1200,
            "line_close": True,
            "bgcolor": "white",
            "fill": "toself",
            "scattersize": 8,
            "connectgaps": True,
            "opacity": 0.5,
            "linewidth": 2,
            "radialaxis": {
                "showline": False,
                "linewidth": 2,
                "linecolor": "rgba(0, 0, 0, 0.25)",
                "gridcolor": "rgba(0, 0, 0, 0.25)",
                "ticks": "outside",
                "tickfont": {"size": 14, "color": "black"},
            },
            "angularaxis": {
                "showline": True,
                "linewidth": 2,
                "linecolor": "rgba(0, 0, 0, 0.25)",
                "gridcolor": "rgba(0, 0, 0, 0.25)",
                "tickfont": {"size": 16, "color": "black"},
            },
            "color_discrete_map": {
                "High Amplitude": "rgba(255, 0, 0, 1)",
                "Low Amplitude": "rgba(0, 0, 255, 1)",
            },
            "title_font": {"family": "Times New Roman", "size": 30, "color": "black"},
            "title_x": 0.5,
            "title_y": None,
            "legend": {
                "yanchor": "top",
                "xanchor": "left",
                "y": 0.99,
                "x": 0.01,
                "title_font_family": "Times New Roman",
                "font": {"size": 12, "color": "black"},
            },
            "mode": "markers+lines",
            "engine": "kaleido",
        }

    @staticmethod
    def transition_matrix() -> dict:
        return _PlotDefaults.caps2corr()


class _PlotFuncs:
    """Helper functions for plotting."""

    @staticmethod
    def base_kwargs(plot_dict: dict, line: bool = True, edge: bool = True) -> dict:
        kwargs = {
            "cmap": plot_dict["cmap"],
            "cbar_kws": {"shrink": plot_dict["shrink"]},
            "annot": plot_dict["annot"],
            "annot_kws": plot_dict["annot_kws"],
            "fmt": plot_dict["fmt"],
            "alpha": plot_dict["alpha"],
            "vmin": plot_dict["vmin"],
            "vmax": plot_dict["vmax"],
        }

        if line:
            kwargs.update(
                {"linewidths": plot_dict["linewidths"], "linecolor": plot_dict["linecolor"]}
            )
        if edge:
            kwargs.update({"edgecolors": plot_dict["edgecolors"]})

        return kwargs

    @staticmethod
    def extra_kwargs(hemisphere_labels: bool) -> tuple[bool, bool]:
        # Determine whether to add line related and edge related kwargs to base kwargs determined
        # by hemisphere_labels
        return (False, False) if hemisphere_labels else (True, True)

    @staticmethod
    def border(
        display: Union[Axes, Figure],
        plot_dict: dict,
        axhline: int,
        axvline: Union[int, None] = None,
    ) -> Union[Axes, Figure]:
        if not plot_dict["borderwidths"]:
            return display

        display.axhline(y=0, color=plot_dict["linecolor"], linewidth=plot_dict["borderwidths"])
        display.axhline(
            y=axhline, color=plot_dict["linecolor"], linewidth=plot_dict["borderwidths"]
        )
        display.axvline(x=0, color=plot_dict["linecolor"], linewidth=plot_dict["borderwidths"])

        if axvline:
            display.axvline(
                x=axvline, color=plot_dict["linecolor"], linewidth=plot_dict["borderwidths"]
            )
        else:
            display.axvline(
                x=axhline, color=plot_dict["linecolor"], linewidth=plot_dict["borderwidths"]
            )

        return display

    @staticmethod
    def label_size(
        display: Union[Axes, Figure], plot_dict: dict, set_x: bool = True, set_y: bool = True
    ) -> Union[Axes, Figure]:
        if set_x:
            display.set_xticklabels(
                display.get_xticklabels(),
                size=plot_dict["xticklabels_size"],
                rotation=plot_dict["xlabel_rotation"],
            )

        if set_y:
            display.set_yticklabels(
                display.get_yticklabels(),
                size=plot_dict["yticklabels_size"],
                rotation=plot_dict["ylabel_rotation"],
            )

        if plot_dict["cbarlabels_size"]:
            cbar = display.collections[0].colorbar
            cbar.ax.tick_params(labelsize=plot_dict["cbarlabels_size"])

        return display

    @staticmethod
    def set_ticks(display: Union[Axes, Figure], labels: list) -> Union[Axes, Figure]:
        ticks = [i for i, label in enumerate(labels) if label]

        display.set_xticks(ticks)
        display.set_xticklabels([label for label in labels if label])
        display.set_yticks(ticks)
        display.set_yticklabels([label for label in labels if label])

        return display

    @staticmethod
    def division_line(
        display: Union[Axes, Figure],
        linewidths: int,
        n_labels: int,
        set_x: bool = True,
        add_labels: Union[list[str], bool] = None,
    ) -> tuple[Union[Axes, Figure], int, int]:
        division_line = n_labels // 2
        left_hemisphere_tick = (0 + division_line) // 2
        right_hemisphere_tick = (division_line + n_labels) // 2

        if set_x:
            display.set_xticks([left_hemisphere_tick, right_hemisphere_tick])

        display.set_yticks([left_hemisphere_tick, right_hemisphere_tick])

        if add_labels:
            display.set_xticklabels(["LH", "RH"])
            display.set_yticklabels(["LH", "RH"])

        line_widths = linewidths if linewidths != 0 else 1

        return display, division_line, line_widths

    @staticmethod
    def set_title(
        display: Union[Axes, Figure],
        title: str,
        suffix: Union[str, None],
        plot_dict: dict,
        is_subplot: bool = False,
    ) -> Union[Axes, Figure]:
        title = f"{title} {suffix}" if suffix else title

        if is_subplot:
            kwarg = {"fontsize": plot_dict["suptitle_fontsize"]}
        elif "fontdict" in inspect.signature(display.set_title).parameters.keys():
            kwarg = {"fontdict": {"fontsize": plot_dict["fontsize"]}}
        else:
            kwarg = {"fontsize": plot_dict["fontsize"]}

        display.suptitle(title, **kwarg) if is_subplot else display.set_title(title, **kwarg)

        return display

    @staticmethod
    def save_fig(
        fig: Union[Axes, Figure], output_dir: str, filename: str, plot_dict: dict, as_pickle: bool
    ) -> None:
        if as_pickle:
            # Allow Axes or Figure
            io_utils._serialize(fig, output_dir, filename.replace(".png", ".pkl"))
        else:
            fig = fig.get_figure() if not hasattr(fig, "savefig") else fig
            fig.savefig(
                os.path.join(output_dir, filename),
                dpi=plot_dict["dpi"],
                bbox_inches=plot_dict["bbox_inches"],
            )

    @staticmethod
    def show(show_figs: bool) -> None:
        plt.show() if show_figs else plt.close("all")


class _MatrixVisualizer:
    """
    Generates heatmaps and saves contents for correlation (``CAP.caps2corr``) and transition
    probability matrices.
    """

    @staticmethod
    def create_display(
        df: DataFrame, plot_dict: dict, suffix_title: str, group: str, call: str
    ) -> Union[Axes, Figure]:
        # Refresh grid for each iteration
        plt.figure(figsize=plot_dict["figsize"])

        display = seaborn.heatmap(
            df, xticklabels=True, yticklabels=True, **_PlotFuncs.base_kwargs(plot_dict)
        )

        # Add Border; returns display if border in `plot_dict` is Falsy
        display = _PlotFuncs.border(display, plot_dict, df.shape[1], df.shape[0])

        # Modify label sizes
        display = _PlotFuncs.label_size(display, plot_dict)

        if call == "trans":
            display.set_ylabel("From", fontdict={"fontsize": plot_dict["fontsize"]})
            display.set_xlabel("To", fontdict={"fontsize": plot_dict["fontsize"]})

        # Set plot name
        plot_name = "Correlation Matrix" if call == "corr" else "Transition Probabilities"
        display = _PlotFuncs.set_title(
            display, f"{group} CAPs {plot_name}", suffix_title, plot_dict
        )

        return display

    @staticmethod
    def save_contents(
        output_dir: str,
        suffix_filename: str,
        group: str,
        curr_dict: dict,
        plot_dict: dict,
        save_plots: bool,
        save_df: bool,
        display: Union[Axes, Figure],
        as_pickle: bool,
        call: str,
    ) -> None:
        # Save figure
        if output_dir:
            io_utils._makedir(output_dir)

            desc = "correlation_matrix" if call == "corr" else "transition_probability_matrix"
            filename = io_utils._filename(f"{group}_CAPs_{desc}", suffix_filename, "suffix", "png")
            if save_plots:
                _PlotFuncs.save_fig(display, output_dir, filename, plot_dict, as_pickle)

            if save_df:
                filename = filename.replace(".png", ".csv")
                curr_dict[group].to_csv(
                    path_or_buf=os.path.join(output_dir, filename), sep=",", index=True
                )



"""A class which is responsible for accessing attributes in ``CAP``."""

import copy, sys
from typing import Union

import numpy as np
from numpy.typing import NDArray
from sklearn.cluster import KMeans

from ..check_parcel_approach import _check_parcel_approach
from ...typing import ParcelConfig, ParcelApproach


class _CAPGetter:
    def __init__(self):
        pass

    ### Attributes exist when CAP initialized
    @property
    def parcel_approach(self) -> Union[ParcelApproach, None]:
        return self._parcel_approach

    @parcel_approach.setter
    def parcel_approach(self, parcel_dict: Union[ParcelConfig, ParcelApproach, str]) -> None:
        self._parcel_approach = _check_parcel_approach(parcel_approach=parcel_dict, call="setter")

    @property
    def groups(self) -> Union[dict[str, list[str]], None]:
        return copy.deepcopy(self._groups)

    ### Attributes exist when CAP.get_caps() used
    @property
    def subject_table(self) -> Union[dict[str, str], None]:
        return copy.deepcopy(getattr(self, "_subject_table", None))

    @subject_table.setter
    def subject_table(self, subject_dict: dict[str, str]) -> None:
        if isinstance(subject_dict, dict):
            self._subject_table = copy.deepcopy(subject_dict)
        else:
            raise TypeError(
                "`self.subject_table` must be a dictionary where the keys are the subject IDs and "
                "the values are the group names."
            )

    @property
    def n_clusters(self) -> Union[int, list[int], None]:
        return getattr(self, "_n_clusters", None)

    @property
    def n_cores(self) -> Union[int, None]:
        return getattr(self, "_n_cores", None)

    @property
    def runs(self) -> Union[list[Union[int, str]], None]:
        return getattr(self, "_runs", None)

    @property
    def standardize(self) -> Union[bool, None]:
        return getattr(self, "_standardize", None)

    @property
    def concatenated_timeseries(self) -> Union[dict[str, NDArray[np.floating]], None]:
        return getattr(self, "_concatenated_timeseries", None)

    @concatenated_timeseries.deleter
    def concatenated_timeseries(self) -> None:
        del self._concatenated_timeseries

    @property
    def means(self) -> Union[dict[str, NDArray[np.floating]], None]:
        return copy.deepcopy(getattr(self, "_mean_vec", None))

    @property
    def stdev(self) -> Union[dict[str, NDArray[np.floating]], None]:
        return copy.deepcopy(getattr(self, "_stdev_vec", None))

    @property
    def kmeans(self) -> Union[dict[str, KMeans], None]:
        return copy.deepcopy(getattr(self, "_kmeans", None))

    @property
    def caps(self) -> Union[dict[str, dict[str, NDArray[np.floating]]], None]:
        return copy.deepcopy(getattr(self, "_caps", None))

    @property
    def cluster_scores(self) -> Union[dict[str, Union[str, dict[str, float]]], None]:
        return getattr(self, "_cluster_scores", None)

    @property
    def cluster_selection_method(self) -> Union[str, None]:
        attr = getattr(self, "_cluster_scores", None)

        if attr:
            return attr["Cluster_Selection_Method"]
        else:
            return attr

    @property
    def optimal_n_clusters(self) -> Union[dict[str, int], None]:
        return getattr(self, "_optimal_n_clusters", None)

    @property
    def variance_explained(self) -> Union[dict[str, float], None]:
        return getattr(self, "_variance_explained", None)

    # Generated in `caps2plot`
    @property
    def region_means(
        self,
    ) -> Union[dict[str, dict[str, Union[list[str], NDArray[np.floating]]]], None]:
        return getattr(self, "_region_means", None)

    @property
    def outer_products(self) -> Union[dict[str, dict[str, NDArray[np.floating]]], None]:
        return getattr(self, "_outer_products", None)

    # Generated in `caps2radar`
    @property
    def cosine_similarity(
        self,
    ) -> Union[dict[str, dict[str, Union[list[str], NDArray[np.floating]]]], None]:
        return getattr(self, "_cosine_similarity", None)

    def _concatenated_timeseries_size(self) -> str:
        if not self.concatenated_timeseries:
            return "0 bytes"

        total_bytes = sum(arr.nbytes for arr in self.concatenated_timeseries.values())
        total_bytes += sys.getsizeof(self.concatenated_timeseries)

        return f"{total_bytes} bytes"

    def __str__(self) -> str:
        """
        Print Current Object State.

        Provides a formatted summary of the ``CAP`` configuration when called with ``print(self)``.
        Returns a string containing the following information:

        - Parcellation approach used
        - Group definitions
        - Clustering configuration (e.g, number of clusters, selection method, etc)
        - Optimal number of clusters  per group (if a range of clusters were provided)
        - Number of CPU cores used for clustering (multiprocessing)
        - Run identifiers used in analysis
        - Estimated memory usage estimate for ``concatenated_timeseries`` (in bytes)
        - Standardization applied prior to clustering
        - Co-Activation Patterns (CAPs) per group
        - Variance explained by clustering

        Returns
        -------
        str
            A formatted string containing information about the object's current state.

        Example
        -------

        >>> cap_analysis = CAP()
        >>> print(cap_analysis)
        Current Object State:
        =====================
        Parcellation Approach                                       : None
        ...
        """
        parcellation_name = list(self.parcel_approach)[0] if self.parcel_approach else None
        # Get group names
        groups_names = ", ".join(f"{k}" for k in self.groups) if self.groups else None
        # Get total CAPs per group
        group_caps = {k: len(v) for k, v in self.caps.items()} if self.caps else None
        # Get additional information
        method = self.cluster_selection_method
        optimal_n = self.optimal_n_clusters
        data_size = self._concatenated_timeseries_size()
        var_explained = self.variance_explained

        object_properties = (
            f"Parcellation Approach                                       : {parcellation_name}\n"
            f"Groups                                                      : {groups_names}\n"
            f"Number of Clusters                                          : {self.n_clusters}\n"
            f"Cluster Selection Method                                    : {method}\n"
            f"Optimal Number of Clusters (if Range of Clusters Provided)  : {optimal_n}\n"
            f"CPU Cores Used for Clustering (Multiprocessing)             : {self.n_cores}\n"
            f"User-Specified Runs IDs Used for Clustering                 : {self.runs}\n"
            f"Concatenated Timeseries Bytes                               : {data_size}\n"
            f"Standardized Concatenated Timeseries                        : {self.standardize}\n"
            f"Co-Activation Patterns (CAPs)                               : {group_caps}\n"
            f"Variance Explained by Clustering                            : {var_explained}"
        )

        sep = "=" * len(object_properties.rsplit(": ")[0])

        return "Current Object State:\n" + sep + f"\n{object_properties}"



from .check_kwargs import _check_kwargs
from .check_parcel_approach import _check_parcel_approach, _collapse_aal_node_names
from .logger import _logger
from .analysis import _CAPGetter, _build_tree, _cap2statmap, _get_target_indices, _run_kmeans
from .extraction import (
    _TimeseriesExtractorGetter,
    _check_confound_names,
    _extract_timeseries,
    _standardize,
)
from .plotting_utils import _MatrixVisualizer, _PlotDefaults, _PlotFuncs



"""Function to standardize timeseries within subject runs."""

from typing import Union, Optional

import neurocaps._utils.io as io_utils
from ..typing import SubjectTimeseries
from .._utils import _standardize


def standardize(
    subject_timeseries_list: Union[list[SubjectTimeseries], list[str]],
    return_dicts: bool = True,
    output_dir: Optional[str] = None,
    filenames: Optional[list[str]] = None,
) -> Union[dict[str, SubjectTimeseries], None]:
    """
    Perform Participant-wise Timeseries Standardization Within Runs.

    Standardizes the columns/ROIs of each run independently for all subjects in the subject
    timeseries. Uses sample standard deviation with Bessel's correction (`n-1` in denominator).
    Primarily to be used when standardizing was not done in ``TimeseriesExtractor``.

    .. note:: Standard deviations below ``np.finfo(std.dtype).eps`` are replaced with 1 for\
    numerical stability.

    Parameters
    ----------
    subject_timeseries_list: :obj:`list[SubjectTimeseries]` or :obj:`list[str]`
        A list where each element consist of a dictionary mapping subject IDs to their run IDs and
        associated timeseries (TRs x ROIs) as a NumPy array. Can also be a list consisting of paths
        to pickle files containing this same structure. Refer to documentation for
        ``SubjectTimeseries`` in the "See Also" section for an example structure.

    return_dicts: :obj:`bool`, default=True
        If True, returns a single dictionary containing the standardized input dictionaries. Keys
        are named "dict_{0}" where {0} corresponds to the dictionary's position in the input list.

    output_dir: :obj:`str` or :obj:`None`, default=None
        Directory to save the standardized dictionaries as pickle files. The directory will be
        created if it does not exist. Dictionaries will not be saved if None.

    filenames: :obj:`list[str]` or :obj:`None`, default=None
        A list of names to save the standardized dictionaries as. Names are matched to dictionaries
        by position (e.g., a file name in the 0th position will be the file name for the dictionary
        in the 0th position of ``subject_timeseries_list``). If None and ``output_dir`` is specified,
        uses default file names - "subject_timeseries_{0}_standardized.pkl" (where {0} indicates the
        original input order).

    Returns
    -------
    dict[str, SubjectTimeseries]
        A nested dictionary containing the standardized subject timeseries if ``return_dicts`` is
        True.

    See Also
    --------
    :data:`neurocaps.typing.SubjectTimeseries`
        Type definition for the subject timeseries dictionary structure.
    """
    assert (
        isinstance(subject_timeseries_list, list) and len(subject_timeseries_list) > 0
    ), "`subject_timeseries_list` must be a list greater than length 0."

    io_utils._issue_file_warning("filenames", filenames, output_dir)

    standardized_dicts = {}

    for indx, curr_dict in enumerate(subject_timeseries_list):
        curr_dict = io_utils._get_obj(curr_dict)

        for subj_id in curr_dict:
            for run in curr_dict[subj_id]:
                curr_dict[subj_id][run] = _standardize(curr_dict[subj_id][run])

        standardized_dicts[f"dict_{indx}"] = curr_dict

    if output_dir:
        io_utils._dicts_to_pickles(
            output_dir=output_dir,
            dict_list=standardized_dicts,
            filenames=filenames,
            call="standardize",
        )

    if return_dicts:
        return standardized_dicts
